# this module contains the http request mechanism

from typing import Any
import urllib.request
import aiohttp
import os.path
import ssl

from .basehttp import HTTP as BaseHTTP
from .basehttp import ClientResponse as BaseClientResponse

from logging import getLogger

log = getLogger(__name__)

# prepare LindenLab.crt
__llcacrt = "LindenLab.crt"
_sslcontext = None
if not _sslcontext:
    if not os.path.isfile("LindenLab.crt"):
        log.debug("Donwload LindenLab.crt")
        urllib.request.urlretrieve(
            "https://raw.githubusercontent.com/secondlife/llca/master/LindenLab.crt",
            __llcacrt,
        )
    _sslcontext = ssl.create_default_context(cafile=__llcacrt)
    _sslcontext.set_ciphers("ALL:@SECLEVEL=1")


# wrap aiohttp.ClientResponse
class ClientResponse(BaseClientResponse):
    __resp: aiohttp.ClientResponse

    def __init__(self, resp: aiohttp.ClientResponse) -> None:
        self.__resp = resp
        self.headers = dict(resp.headers)

    async def text(self) -> str:
        return await self.__resp.text()

    @property
    def status(self) -> int:
        return self.__resp.status

    @property
    def reason(self) -> str | None:
        return self.__resp.reason


# default HTTP implementation, provides by dependency-injector
class HTTP(BaseHTTP):
    # http get method
    @staticmethod
    async def get(url: str, headers: dict[str, Any] = dict()) -> ClientResponse:
        log.debug(f"get: {url=}")
        async with aiohttp.ClientSession() as session:
            async with session.get(url, ssl=_sslcontext, headers=headers) as resp:
                log.debug(f"{resp}{await resp.text()}")
                if resp.status not in range(200, 203):
                    e = await HTTP.__exceptionByResp(resp)
                    log.error(e)
                    raise e
                return ClientResponse(resp)

    # http post method
    @staticmethod
    async def post(
        url: str, data: str | None, headers: dict[str, Any] = dict()
    ) -> ClientResponse:
        log.debug(f"post: {url=}; {data=}")
        async with aiohttp.ClientSession() as session:
            async with session.post(
                url, data=data, ssl=_sslcontext, headers=headers
            ) as resp:
                await resp.text()
                if resp.status not in range(200, 203):
                    e = await HTTP.__exceptionByResp(resp)
                    log.error(e)
                    raise e
                return ClientResponse(resp)



from pydantic import BaseModel
from uuid import UUID
import re

from lslgwlib.models import HTTPData, Avatar, Region


# LinkSet API responses model
class LSLResponse(HTTPData):
    def __init__(self, *args, **kwargs) -> None:
        # constructor by http response and parsed data
        if (
            len(args) == 2
            and hasattr(args[0], "headers")
            and isinstance(args[1], BaseModel | list | int | float | str | UUID | None)
        ):
            headers: dict[str, str] = dict()
            for headName in args[0].headers:
                if not headName.startswith("X-SecondLife"):
                    headers[headName] = args[0].headers[headName]
            super().__init__(
                owner=Avatar(
                    args[0].headers["X-SecondLife-Owner-Key"],
                    args[0].headers["X-SecondLife-Owner-Name"],
                ),
                objectKey=UUID(args[0].headers["X-SecondLife-Object-Key"]),
                objectName=args[0].headers["X-SecondLife-Object-Name"],
                position=re.sub(
                    r"(\(|\))", "", args[0].headers["X-SecondLife-Local-Position"]
                ).split(","),
                rotation=re.sub(
                    r"(\(|\))", "", args[0].headers["X-SecondLife-Local-Rotation"]
                ).split(","),
                velocity=re.sub(
                    r"(\(|\))", "", args[0].headers["X-SecondLife-Local-Velocity"]
                ).split(","),
                region=Region(args[0].headers["X-SecondLife-Region"]),
                production=args[0].headers["X-SecondLife-Shard"] == "Production",
                data=args[1],
                headers=headers,
            )
        else:
            super().__init__(**kwargs)



from dependency_injector.wiring import Provide, inject
from typing import Any

from .container import Container
from .linkset import LinkSet
from .basehttp import HTTP


class API:
    """lsl-gateway-client base class"""

    container = Container()

    def __init__(self) -> None:
        # wiring dependency-injector
        self.container.wire(modules=[__name__])

    @inject
    def linkset(
        self,
        url: str,
        headers: dict[str, Any] = dict(),
        http: HTTP = Provide[Container.http],
    ) -> LinkSet:
        """get LinkSet, it provides LSL object iteraction

        Arguments:
        url -     link to LSL object
        headers - optional, add headers for any request
        """
        return LinkSet(http, url, headers)



# this module contains the http request mechanism
# fake for unit tests

import abc
from typing_extensions import Annotated, Any
import aiohttp.web_exceptions as excepts
from pydantic import validate_call, Field


class ClientResponse(abc.ABC):
    """HTTP response representation"""

    headers: dict[str, str]

    @abc.abstractmethod
    def __init__(self, text: str) -> None:
        raise NotImplementedError()

    @abc.abstractmethod
    async def text(self) -> str:
        """Response body"""
        raise NotImplementedError()

    @property
    def status(self) -> int:
        """Response status code"""
        raise NotImplementedError()

    @property
    def reason(self) -> str | None:
        """Response error reason"""
        raise NotImplementedError()


class HTTP:
    """HTTP client"""

    @staticmethod
    @abc.abstractmethod
    async def get(url: str, headers: dict[str, Any] = dict()) -> ClientResponse:
        """HTTP GET

        Arguments:
        url - adress with query string
        """
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    async def post(
        url: str, data: str | None, headers: dict[str, Any] = dict()
    ) -> ClientResponse:
        """HTTP POST

        Arguments:
        url -  adress with query string
        data - body
        """
        raise NotImplementedError()

    # select exception type by responce.status
    @staticmethod
    @validate_call
    async def __exceptionByResp(
        resp=Annotated[ClientResponse, Field()]
    ) -> excepts.HTTPException:
        """Make aiohttp exception from ClientResponse

        Arguments:
        resp - ClientResponse
        """
        code = resp.status
        text = await resp.text()
        if text:
            reason = text
        elif resp.reason:
            reason = resp.reason
        else:
            reason = None
        err: Exception = excepts.HTTPError(reason=reason, headers=resp.headers)
        match code:
            case 200:
                err = excepts.HTTPOk(reason=reason, headers=resp.headers)
            case 201:
                err = excepts.HTTPCreated(reason=reason, headers=resp.headers)
            case 202:
                err = excepts.HTTPAccepted(reason=reason, headers=resp.headers)
            case 203:
                err = excepts.HTTPNonAuthoritativeInformation(
                    reason=reason, headers=resp.headers
                )
            case 204:
                err = excepts.HTTPNoContent(reason=reason, headers=resp.headers)
            case 205:
                err = excepts.HTTPResetContent(reason=reason, headers=resp.headers)
            case 206:
                err = excepts.HTTPPartialContent(reason=reason, headers=resp.headers)

            case 300:
                err = excepts.HTTPMultipleChoices(
                    reason=reason,
                    headers=resp.headers,
                    location=resp.headers["Location"],
                )
            case 301:
                err = excepts.HTTPMovedPermanently(
                    reason=reason,
                    headers=resp.headers,
                    location=resp.headers["Location"],
                )
            case 302:
                err = excepts.HTTPFound(
                    reason=reason,
                    headers=resp.headers,
                    location=resp.headers["Location"],
                )
            case 303:
                err = excepts.HTTPSeeOther(
                    reason=reason,
                    headers=resp.headers,
                    location=resp.headers["Location"],
                )
            case 304:
                err = excepts.HTTPNotModified(reason=reason, headers=resp.headers)
            case 305:
                err = excepts.HTTPUseProxy(
                    reason=reason,
                    headers=resp.headers,
                    location=resp.headers["Location"],
                )
            case 307:
                err = excepts.HTTPTemporaryRedirect(
                    reason=reason,
                    headers=resp.headers,
                    location=resp.headers["Location"],
                )
            case 308:
                err = excepts.HTTPPermanentRedirect(
                    reason=reason,
                    headers=resp.headers,
                    location=resp.headers["Location"],
                )

            case 400:
                err = excepts.HTTPBadRequest(reason=reason, headers=resp.headers)
            case 401:
                err = excepts.HTTPUnauthorized(reason=reason, headers=resp.headers)
            case 402:
                err = excepts.HTTPPaymentRequired(reason=reason, headers=resp.headers)
            case 403:
                err = excepts.HTTPForbidden(reason=reason, headers=resp.headers)
            case 404:
                err = excepts.HTTPNotFound(reason=reason, headers=resp.headers)
            case 405:
                err = excepts.HTTPMethodNotAllowed(
                    reason=reason,
                    headers=resp.headers,
                    method=resp.method,
                    allowed_methods=resp.headers["Allow"].split(),
                )
            case 406:
                err = excepts.HTTPNotAcceptable(reason=reason, headers=resp.headers)
            case 407:
                err = excepts.HTTPProxyAuthenticationRequired(
                    reason=reason, headers=resp.headers
                )
            case 408:
                err = excepts.HTTPRequestTimeout(reason=reason, headers=resp.headers)
            case 409:
                err = excepts.HTTPConflict(reason=reason, headers=resp.headers)
            case 410:
                err = excepts.HTTPGone(reason=reason, headers=resp.headers)
            case 411:
                err = excepts.HTTPLengthRequired(reason=reason, headers=resp.headers)
            case 412:
                err = excepts.HTTPPreconditionFailed(
                    reason=reason, headers=resp.headers
                )
            case 413:
                err = excepts.HTTPRequestEntityTooLarge(
                    reason=reason,
                    headers=resp.headers,
                    max_size=0,
                    actual_size=resp.headers["Content-Length"],
                )
            case 414:
                err = excepts.HTTPRequestURITooLong(reason=reason, headers=resp.headers)
            case 415:
                err = excepts.HTTPUnsupportedMediaType(
                    reason=reason, headers=resp.headers
                )
            case 416:
                err = excepts.HTTPRequestRangeNotSatisfiable(
                    reason=reason, headers=resp.headers
                )
            case 417:
                err = excepts.HTTPExpectationFailed(reason=reason, headers=resp.headers)
            case 421:
                err = excepts.HTTPMisdirectedRequest(
                    reason=reason, headers=resp.headers
                )
            case 422:
                err = excepts.HTTPUnprocessableEntity(
                    reason=reason, headers=resp.headers
                )
            case 424:
                err = excepts.HTTPFailedDependency(reason=reason, headers=resp.headers)
            case 426:
                err = excepts.HTTPUpgradeRequired(reason=reason, headers=resp.headers)
            case 429:
                err = excepts.HTTPTooManyRequests(reason=reason, headers=resp.headers)
            case 431:
                err = excepts.HTTPRequestHeaderFieldsTooLarge(
                    reason=reason, headers=resp.headers
                )
            case 451:
                err = excepts.HTTPUnavailableForLegalReasons(
                    reason=reason, headers=resp.headers, link=None
                )

            case 500:
                err = excepts.HTTPInternalServerError(
                    reason=reason, headers=resp.headers
                )
            case 501:
                err = excepts.HTTPNotImplemented(reason=reason, headers=resp.headers)
            case 502:
                err = excepts.HTTPBadGateway(reason=reason, headers=resp.headers)
            case 503:
                err = excepts.HTTPServiceUnavailable(
                    reason=reason, headers=resp.headers
                )
            case 504:
                err = excepts.HTTPGatewayTimeout(reason=reason, headers=resp.headers)
            case 505:
                err = excepts.HTTPVersionNotSupported(
                    reason=reason, headers=resp.headers
                )
            case 506:
                err = excepts.HTTPVariantAlsoNegotiates(
                    reason=reason, headers=resp.headers
                )
            case 507:
                err = excepts.HTTPInsufficientStorage(
                    reason=reason, headers=resp.headers
                )
            case 510:
                err = excepts.HTTPNotExtended(reason=reason, headers=resp.headers)
            case 511:
                err = excepts.HTTPNetworkAuthenticationRequired(
                    reason=reason, headers=resp.headers
                )

            case code if code < 300:
                err = excepts.HTTPSuccessful(reason=reason, headers=resp.headers)
            case code if code < 400:
                err = excepts.HTTPRedirection(reason=reason, headers=resp.headers)
            case code if code < 500:
                err = excepts.HTTPClientError(reason=reason, headers=resp.headers)
            case _:
                err = excepts.HTTPServerError(reason=reason, headers=resp.headers)

        return err



from typing_extensions import Annotated, Any
from pydantic import validate_call, Field
from uuid import UUID, uuid5
from typing import Sequence
import re

from lslgwclient.models import LSLResponse
from lslgwclient.exceptions import linksetDataExceptionByNum
from lslgwlib.models import (
    LinkSetInfo,
    PrimInfo,
    Avatar,
    Permissions,
    Invetory,
    InvetoryItem,
)
from lslgwlib.enums import InvetoryType
from .basehttp import HTTP

from logging import getLogger, Logger


# provides API for server.lsl inworld
class LinkSet:
    """LinkSet - provides LSL object interaction"""

    __log: Logger = getLogger()
    __urlPattern = re.compile(
        r"^https://[-a-z0-9@:%_\+~#=]{1,255}\.agni\.secondlife\.io:12043/cap/"
        + r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$",
        re.IGNORECASE,
    )
    __headers: dict[str, Any]
    __http: HTTP
    __url: str
    __id: UUID

    # contructor by LSLHttp url
    def __init__(self, http: HTTP, url: str, headers: dict[str, Any] = dict()) -> None:
        if not self.__urlPattern.match(url):
            raise ValueError(f"Invalid url: {url}")
        self.__headers = headers
        self.__url = url.lower()
        self.__http = http
        self.__log.info(url)

    # API info method
    async def info(self) -> LSLResponse:
        """LSL object base stats

        returns LSLResponse.data: LinkSetInfo
        """
        resp = await self.__http.get(f"{self.__url}/info", headers=self.__headers)
        body = (await resp.text()).split("¦")
        lslresp = LSLResponse(
            resp,
            LinkSetInfo(
                id=resp.headers["X-SecondLife-Object-Key"],
                owner=Avatar(
                    resp.headers["X-SecondLife-Owner-Key"],
                    resp.headers["X-SecondLife-Owner-Name"],
                ),
                lastOwnerId=UUID(body[0]),
                creatorId=UUID(body[1]),
                groupId=UUID(body[2]),
                name=resp.headers["X-SecondLife-Object-Name"],
                description=body[3],
                attached=body[4],
                primsNum=body[5],
                inventoryNum=body[6],
                createdAt=body[7],
                rezzedAt=body[8],
                scriptName=body[9],
                permissions=Permissions(*body[10:]),
            ),
        )
        self.__id = lslresp.data.id
        self.__log = getLogger(f"{self.__class__.__name__}({self.__id})")
        self.__log.debug(lslresp.data)
        return lslresp

    # API prims method
    async def prims(self) -> LSLResponse:
        """LSL object linkset prims

        returns LSLResponse.data: list[PrimInfo]
        """
        # list of downloaded prims info
        prims: list[PrimInfo] = list()
        # already used ids (for exclude doubles)
        ids: list[UUID] = list()

        # converts string returned by server.lsl to PrimInfo model
        def primInfo(info) -> PrimInfo:
            # gen unique id for every prim in linkset
            def primId(creator: str, created: str) -> UUID:
                tmpId = uuid5(UUID(creator), created)
                n = 0
                while tmpId in ids:
                    tmpId = uuid5(UUID(creator), f"{created}{n}")
                    n += 1
                ids.append(tmpId)
                return tmpId

            pId = primId(info[0], info[3] + info[4])
            return PrimInfo(
                id=pId,
                creatorId=info[0],
                createdAt=info[3],
                name=info[1],
                description=info[2],
                faces=info[4],
            )

        body = ["+"]

        # load parts while exists
        while body and body[-1] == "+":
            if len(prims):
                resp = await self.__http.get(
                    f"{self.__url}/prims/{len(prims)+1}", headers=self.__headers
                )
            else:
                resp = await self.__http.get(
                    f"{self.__url}/prims", headers=self.__headers
                )
            body = (await resp.text()).splitlines()
            for line in body:
                if line != "+":
                    prims.append(primInfo(line.split("¦")))

        self.__log.debug(f"{len(prims)} prims")
        return LSLResponse(resp, prims)

    # get all linkset data keys
    async def linksetDataKeys(self) -> LSLResponse:
        """LSL object likset data keys

        returns LSLResponse.data: list[str]
        """
        keys: list[str] = list()
        body = ["+"]

        while body[-1] == "+":
            if len(keys):
                resp = await self.__http.get(
                    f"{self.__url}/linksetdata/keys/{len(keys)+1}",
                    headers=self.__headers,
                )
            else:
                resp = await self.__http.get(
                    f"{self.__url}/linksetdata/keys", headers=self.__headers
                )
            body = (await resp.text()).split("¦")
            if body[-1] == "+":
                keys.extend(body[:-1])
            else:
                keys.extend(body)

        self.__log.debug(f"{len(keys)} linkset data keys")
        return LSLResponse(resp, keys)

    # get linkset data value by key
    @validate_call
    async def linksetDataGet(
        self, key: Annotated[str, Field(min_length=1)], pwd: str | None = None
    ) -> LSLResponse:
        """LSL object likset data get value by key

        returns LSLResponse.data: str
        """
        if pwd:
            resp = await self.__http.post(
                f"{self.__url}/linksetdata/read/{key}", pwd, headers=self.__headers
            )
        else:
            resp = await self.__http.get(
                f"{self.__url}/linksetdata/read/{key}", headers=self.__headers
            )
        if not await resp.text():
            raise linksetDataExceptionByNum(4, key)
        self.__log.debug(f"{key}: {await resp.text()}")
        return LSLResponse(resp, await resp.text())

    # set linkset data value
    @validate_call
    async def linksetDataWrite(
        self,
        key: Annotated[str, Field(min_length=1)],
        value: Annotated[str, Field(min_length=1)],
        pwd: str | None = None,
    ) -> LSLResponse:
        """LSL object likset data set value by key

        Arguments:
        key -   key string
        value - value string
        pwd -   optional protection password
        """
        if pwd:
            resp = await self.__http.post(
                f"{self.__url}/linksetdata/write/{key}",
                f"{value}¦{pwd}",
                headers=self.__headers,
            )
        else:
            resp = await self.__http.post(
                f"{self.__url}/linksetdata/write/{key}", value, headers=self.__headers
            )
        num = int(await resp.text())
        self.__log.debug(f"{key}: {num}")
        if num:
            raise linksetDataExceptionByNum(num, key)
        return LSLResponse(resp, None)

    # delete linkset data value by key
    @validate_call
    async def linksetDataDelete(
        self,
        key: Annotated[str, Field(min_length=1)],
        pwd: str | None = None,
    ) -> LSLResponse:
        """LSL object likset data delete value by key

        Arguments:
        key -   key string
        pwd -   optional protection password
        """
        resp = await self.__http.post(
            f"{self.__url}/linksetdata/delete/{key}", pwd, headers=self.__headers
        )
        num = int(await resp.text())
        self.__log.debug(f"{key}: {num}")
        if num:
            raise linksetDataExceptionByNum(num, key)
        return LSLResponse(resp, None)

    # load inventory
    @validate_call
    async def inventoryRead(
        self, bytype: InvetoryType = InvetoryType.ANY
    ) -> LSLResponse:
        """LSL object inventory get

        Arguments:
        bytype - optional, filter InvetoryType

        returns LSLResponse.data: Invetory
        """
        items: list[InvetoryItem] = list()
        body = ["+"]

        while body[-1] == "+":
            if len(items):
                resp = await self.__http.get(
                    f"{self.__url}/inventory/read/{len(items)}?type={bytype}",
                    headers=self.__headers,
                )
            else:
                resp = await self.__http.get(
                    f"{self.__url}/inventory/read?type={bytype}", headers=self.__headers
                )
            text = await resp.text()
            body = text.split("\n")
            for line in body:
                if line != "+" and len(line):
                    args = line.split("¦")
                    items.append(
                        InvetoryItem(
                            id=args[0],
                            type=args[1],
                            name=args[2],
                            description=args[3],
                            creatorId=args[4],
                            permissions=Permissions(*args[5:10]),
                            acquireTime=args[10],
                        )
                    )

        self.__log.debug(f"{len(items)} inventory items")
        return LSLResponse(resp, Invetory(items=items, filtered=bytype))

    # delete from inventory
    @validate_call
    async def inventoryDelete(
        self, items: Annotated[Sequence[str], Field(min_length=1)]
    ):
        """LSL object inventory delete

        Arguments:
        items - Sequence[str] names
        """
        for item in items:
            if not re.match(r"^[\x20-\x7b\x7d-\x7e]{1,63}$", item):
                raise ValueError(f"'{item}' is not valid item name")

        async def dosend(body: str) -> None:
            await self.__http.post(
                f"{self.__url}/inventory/delete", body, headers=self.__headers
            )
            body = ""

        body = ""
        for item in items:
            if len(body.encode("UTF-8")) + 2 + len(item) > 2048:
                await dosend(body)
                body = ""
            if len(body):
                body += f"¦{item}"
            else:
                body = item
        if len(body):
            await dosend(body)

        self.__log.debug(f"{len(items)} deleted inventory items")

    # give inventory item
    @validate_call
    async def inventoryGive(
        self,
        destination: UUID,
        item: Annotated[str, Field(pattern=r"^[\x20-\x7b\x7d-\x7e]{1,63}$")],
    ):
        """LSL object give item from inventory

        Arguments:
        destination - UUID avatar or object id
        item -        string item name
        """
        if not destination.int:
            raise ValueError("Can't give inventory to NULL_KEY")
        await self.__http.post(
            f"{self.__url}/inventory/give",
            f"{destination}¦{item}",
            headers=self.__headers,
        )

        self.__log.debug(f"Given to {destination} '{item}' inventory item")

    # give inventory items list
    @validate_call
    async def inventoryGiveList(
        self,
        destination: UUID,
        folder: Annotated[str, Field(pattern=r"^[\x20-\x7b\x7d-\x7e]{1,63}$")],
        items: Annotated[Sequence[str], Field(min_length=1, max_length=41)],
    ):
        """LSL object give items from inventory

        Arguments:
        destination - UUID avatar or object id
        item -        string item name
        """
        for item in items:
            if not re.match(r"^[\x20-\x7b\x7d-\x7e]{1,63}$", item):
                raise ValueError(f"'{item}' is not valid item name")

        if not destination.int:
            raise ValueError("Can't give inventory to NULL_KEY")
        body = f'{destination}¦{folder}¦{"¦".join(items)}'
        if len(body.encode("UTF-8")) > 2048:
            raise ValueError("Too big")
        await self.__http.post(
            f"{self.__url}/inventory/givelist", body, headers=self.__headers
        )

        self.__log.debug(f"Given to {destination} '{folder}' inventory items list")



class LinksetDataException(Exception):
    def __init__(self, msg: str):
        super().__init__(msg)


class LinksetDataMemoryException(LinksetDataException):
    def __init__(self, key: str):
        super().__init__(
            f"An error occurred while trying to write linkset data '{key}': not enough memory"
        )


class LinksetDataNoKeyException(LinksetDataException):
    def __init__(self):
        super().__init__(
            "An error occurred while trying to write linkset data: empty key"
        )


class LinksetDataProtectedException(LinksetDataException):
    def __init__(self, key: str):
        super().__init__(
            f"An error occurred while trying to write linkset data '{key}': protected from overwrite"
        )


class LinksetDataNotFoundException(LinksetDataException):
    def __init__(self, key: str):
        super().__init__(
            f"An error occurred while trying to delete linkset data '{key}': not found"
        )


class LinksetDataNotUpdatedException(LinksetDataException):
    def __init__(self, key: str):
        super().__init__(
            f"An error occurred while trying to write linkset data '{key}': not changed"
        )


def exceptionByNum(num: int, msg: str = "") -> LinksetDataException:
    match num:
        case 1:
            return LinksetDataMemoryException(msg)
        case 2:
            return LinksetDataNoKeyException()
        case 3:
            return LinksetDataProtectedException(msg)
        case 4:
            return LinksetDataNotFoundException(msg)
        case 5:
            return LinksetDataNotUpdatedException(msg)
        case _:
            return LinksetDataException(msg)
